/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

#include "operator/operator.pmh"

#define	FE_NOP_DEBUG	FALSE

namespace fe
{

void SurfaceNormalOp::initialize(void)
{
	catalog<String>("icon")="FE_beta";

	catalog<String>("Group");
	catalog<String>("Group","label")="Point Group";
	catalog<String>("Group","suggest")="pointGroups";

	catalog<bool>("PointNormals")=true;
	catalog<String>("PointNormals","label")="Add Point Normals";

	catalog<bool>("PrimitiveNormals")=false;
	catalog<String>("PrimitiveNormals","label")="Add Primitive Normals";

	catalog< sp<Component> >("Input Surface");

	catalog< sp<Component> >("Driver Surface");
	catalog<bool>("Driver Surface","optional")=true;
}

void SurfaceNormalOp::handle(Record& a_rSignal)
{
	const BWORD pointNormals=catalog<bool>("PointNormals");
	const BWORD primitiveNormals=catalog<bool>("PrimitiveNormals");

	if(!pointNormals && !primitiveNormals)
	{
		return;
	}

	const String group=catalog<String>("Group");
	const BWORD grouped=(!group.empty());

	sp<SurfaceAccessibleI> spOutputAccessible;
	if(!accessOutput(spOutputAccessible,a_rSignal)) return;

	sp<SurfaceAccessorI> spOutputProperties;
	if(!access(spOutputProperties,spOutputAccessible,
			e_primitive,e_properties)) return;

	sp<SurfaceI> spDriver;
	access(spDriver,"Driver Surface",e_quiet);
	if(spDriver.isValid())
	{
		spDriver->setRefinement(0);
	}

	sp<SurfaceAccessorI> spOutputVertices;
	if(!access(spOutputVertices,spOutputAccessible,
			e_primitive,e_vertices)) return;

	sp<SurfaceAccessorI> spOutputPoint;
	if(!access(spOutputPoint,spOutputAccessible,e_point,e_position)) return;

	sp<SurfaceAccessorI> spOutputGroup;
	if(grouped)
	{
		if(!access(spOutputGroup,spOutputAccessible,
				e_pointGroup,group)) return;
	}
	else
	{
		spOutputGroup=spOutputPoint;
	}

	const U32 pointCount=spOutputPoint->count();

	const U32 count=spOutputGroup->count();
	if(!count)
	{
		feLog("SurfaceNormalOp::handle point count is zero\n");
		catalog<String>("error")="points required";
		return;
	}

#if	FE_NOP_DEBUG
	feLog("SurfaceNormalOp::handle count %d\n",count);
#endif

	sp<SurfaceAccessorI> spPrimitiveNormal;
	if(primitiveNormals)
	{
		if(!access(spPrimitiveNormal,spOutputAccessible,e_primitive,e_normal,
				e_quiet,e_createMissing)) return;
	}

	sp<SurfaceAccessorI> spPointNormal;
	BWORD inGroup[pointCount];
	if(pointNormals)
	{
		BWORD clearOthers=FALSE;
		if(!grouped || !access(spPointNormal,spOutputAccessible,
				e_point,e_normal,e_quiet))
		{
			clearOthers=grouped;
			if(!access(spPointNormal,spOutputAccessible,e_point,e_normal,
					e_quiet,e_createMissing)) return;
		}

		const SpatialVector zero(0.0,0.0,0.0);

		if(grouped)
		{
			for(U32 pointIndex=0;pointIndex<pointCount;pointIndex++)
			{
				inGroup[pointIndex]=FALSE;
			}
		}

		//* clear normals
		if(clearOthers)
		{
			for(U32 index=0;index<pointCount;index++)
			{
				spPointNormal->set(index,zero);
			}
			for(U32 index=0;index<count;index++)
			{
				inGroup[spOutputGroup->integer(index)]=TRUE;
			}
		}
		else
		{
			for(U32 index=0;index<count;index++)
			{
				const I32 pointIndex=
						grouped? spOutputGroup->integer(index): index;

				inGroup[pointIndex]=TRUE;

				spPointNormal->set(pointIndex,zero);
			}
		}
	}

	//* accumulate face normals to points
	const U32 primitiveCount=spOutputVertices->count();
	for(U32 primitiveIndex=0;primitiveIndex<primitiveCount;primitiveIndex++)
	{
		const U32 subCount=spOutputVertices->subCount(primitiveIndex);
		if(!subCount)
		{
			continue;
		}

		const BWORD openCurve=
				spOutputProperties->integer(primitiveIndex,e_openCurve);
		const U32 countU=
				spOutputProperties->integer(primitiveIndex,e_countU);
		const U32 countV=
				spOutputProperties->integer(primitiveIndex,e_countV);
		const BWORD wrappedU=
				spOutputProperties->integer(primitiveIndex,e_wrappedU);
		const BWORD wrappedV=
				spOutputProperties->integer(primitiveIndex,e_wrappedV);

#if	FE_NOP_DEBUG
		feLog("SurfaceNormalOp::handle primitive %d/%d subCount %d\n",
				primitiveIndex,primitiveCount,subCount);
		feLog("  curve %d mesh %d %d wrapped %d %d\n",
				openCurve,countU,countV,wrappedU,wrappedV);
#endif

		if(openCurve)
		{
			//* open curves

			const SpatialVector rootPoint=
					spOutputVertices->spatialVector(primitiveIndex,0);
			SpatialVector normal(0.0,0.0,1.0);
			if(spDriver.isValid())
			{
				sp<SurfaceI::ImpactI> spImpact=
						spDriver->nearestPoint(rootPoint);
				if(spImpact.isValid())
				{
					normal=spImpact->du();
				}
			}

			if(primitiveNormals)
			{
				spPrimitiveNormal->set(primitiveIndex,unitSafe(normal));
			}
			if(!pointNormals)
			{
				continue;
			}

			const I32 rootIndex=
					spOutputVertices->integer(primitiveIndex,0);
			if(!grouped || inGroup[rootIndex])
			{
				spPointNormal->set(rootIndex,unitSafe(normal));
			}

			if(subCount<2)
			{
				continue;
			}

			SpatialVector lastPoint=rootPoint;
			SpatialVector lastNormal=normal;
			SpatialVector lastSegment;
			Real lastLength;

			SpatialVector thisPoint=
					spOutputVertices->spatialVector(primitiveIndex,1);

			for(U32 subIndex=1;subIndex<subCount;subIndex++)
			{
				const I32 pointIndex=
						spOutputVertices->integer(primitiveIndex,subIndex);
				const BWORD writeNormal=(!grouped || inGroup[pointIndex]);

				if(subIndex==subCount-1)
				{
					if(writeNormal)
					{
						spPointNormal->set(pointIndex,unitSafe(lastNormal));
					}
					continue;
				}

				const SpatialVector nextPoint=spOutputVertices->spatialVector(
						primitiveIndex,subIndex+1);

				//* TODO groups

				SpatialVector nextSegment=nextPoint-thisPoint;
				Real nextLength=magnitude(nextSegment);
				nextSegment*=(nextLength>0.0)? 1.0/nextLength: 1.0;

				if(subIndex==1)
				{
					lastSegment=thisPoint-lastPoint;
					lastLength=magnitude(lastSegment);
					lastSegment*=(lastLength>0.0)? 1.0/lastLength: 1.0;
				}

				SpatialTransform rotation;
				set(rotation,lastSegment,nextSegment);

				SpatialVector nextNormal;
				rotateVector(rotation,lastNormal,nextNormal);

				normal=lastLength*lastNormal+nextLength*nextNormal;

				if(writeNormal)
				{
					spPointNormal->set(pointIndex,unitSafe(normal));
				}

				lastPoint=thisPoint;
				lastNormal=nextNormal;
				lastSegment=nextSegment;
				lastLength=nextLength;

				thisPoint=nextPoint;
			}

			continue;
		}

		//* mesh
		if(countV>1)
		{
			const U32 maxU=wrappedU? countU: countU-1;
			const U32 maxV=wrappedV? countV: countV-1;

			for(U32 v=0;v<maxV;v++)
			{
				const U32 vv=v*countU;
				for(U32 u=0;u<maxU;u++)
				{
					U32 p[4];
					p[0]=vv+u;
					p[1]=(u==countU-1)? vv: p[0]+1;
					p[2]=(v==countV-1)? u: p[0]+countU;
					p[3]=(u==countU-1)? p[2]-(countU-1): p[2]+1;

					SpatialVector sv[4];
					sv[0]=spOutputVertices->spatialVector(primitiveIndex,p[0]);
					sv[1]=spOutputVertices->spatialVector(primitiveIndex,p[1]);
					sv[2]=spOutputVertices->spatialVector(primitiveIndex,p[2]);
					sv[3]=spOutputVertices->spatialVector(primitiveIndex,p[3]);

					SpatialVector faceNormal[2]=
					faceNormal[0]= -unitSafe(cross(sv[1]-sv[0],sv[2]-sv[0]));
					faceNormal[1]= -unitSafe(cross(sv[1]-sv[2],sv[3]-sv[2]));

					if(pointNormals)
					{
						for(U32 half=0;half<2;half++)
						{
							for(U32 m=0;m<3;m++)
							{
								const U32 mm=half+m;
								const I32 pointIndex=spOutputVertices->integer(
										primitiveIndex,p[mm]);
								if(!grouped || inGroup[pointIndex])
								{
									const SpatialVector normal=
											spPointNormal->spatialVector(
											pointIndex);
									spPointNormal->set(pointIndex,
											normal+faceNormal[half]);
#if	FE_NOP_DEBUG
									feLog("    uv %d/%d %d/%d"
											" half %d n %s -> %s\n",
											u,countU,v,countV,half,
											c_print(normal),
											c_print(
											spPointNormal->spatialVector(
											pointIndex)));
#endif
								}
							}
						}
					}

					if(primitiveNormals)
					{
						const SpatialVector normal=
								spPrimitiveNormal->spatialVector(
								primitiveIndex);
						spPrimitiveNormal->set(primitiveIndex,
								normal+faceNormal[0]+faceNormal[1]);
					}
				}
			}
			if(primitiveNormals)
			{
				const SpatialVector normal=
						spPrimitiveNormal->spatialVector(primitiveIndex);
				spPrimitiveNormal->set(primitiveIndex,unitSafe(normal));
			}
			continue;
		}

		if(subCount<3)
		{
			continue;
		}

		//* otherwise, presume triangles

		const SpatialVector p0=
				spOutputVertices->spatialVector(primitiveIndex,0);
		const SpatialVector p1=
				spOutputVertices->spatialVector(primitiveIndex,1);
		const SpatialVector p2=
				spOutputVertices->spatialVector(primitiveIndex,2);

		const SpatialVector faceNormal= -unitSafe(cross(p1-p0,p2-p0));

#if	FE_NOP_DEBUG
		feLog("  prim %d/%d v0 %d v1 %d v2 %d\n",
					primitiveIndex,primitiveCount,
					spOutputVertices->integer(primitiveIndex,0),
					spOutputVertices->integer(primitiveIndex,1),
					spOutputVertices->integer(primitiveIndex,2));
		feLog("  p0 %s p1 %s p2 %s n %s\n",
					c_print(p0),c_print(p1),c_print(p2),c_print(faceNormal));
#endif

		if(pointNormals)
		{
			for(U32 subIndex=0;subIndex<subCount;subIndex++)
			{
				const I32 pointIndex=
						spOutputVertices->integer(primitiveIndex,subIndex);

				if(!grouped || inGroup[pointIndex])
				{
					const SpatialVector normal=
							spPointNormal->spatialVector(pointIndex);
					spPointNormal->set(pointIndex,normal+faceNormal);
#if	FE_NOP_DEBUG
					feLog("    vert %d/%d n %s -> %s\n",
							subIndex,subCount,
							c_print(normal),c_print(
							spPointNormal->spatialVector(pointIndex)));
#endif
				}
			}
		}

		if(primitiveNormals)
		{
			spPrimitiveNormal->set(primitiveIndex,faceNormal);
		}
	}

	//* normalize accumulated normals
	if(pointNormals)
	{
		for(U32 index=0;index<count;index++)
		{
			const I32 pointIndex=grouped? spOutputGroup->integer(index): index;

			const SpatialVector normal=
					spPointNormal->spatialVector(pointIndex);
			spPointNormal->set(pointIndex,unitSafe(normal));
		}
	}
}

} // namespace
